local name, private = ...

local debug = false
function private:debug(...)
    -- Prints a debug message to the chat window.
    -- Will do nothing if debugging is off (see above).
    if debug then
        print(...)
    end
end

if debug then
    print("DEBUGGING IS ON! TURN IT OFF!")
end

------------------------------------------------------------
------------------- The object prototype -------------------
------------------------------------------------------------

local function getNew(self, ...)
    return self:new(...)
end

private.objects = {}
private.objects.__call = getNew

------------------------------------------------------------
---------------------- The list class ----------------------
------------------------------------------------------------

private.objects.list = {}
setmetatable(private.objects.list, private.objects)

function private.objects.list:new(tbl)
    -- Creates a new list, using existing values if provided
    local r = tbl or {}
    setmetatable(r, self)
    self.__index = self
    self.__call = getNew
    return r
end

function private.objects.list:add(item)
    -- Add an item to the list.
    -- Returns true.
    table.insert(self, item)
    return true
end

function private.objects.list:remove(item)
    -- Remove an item from the list.
    -- Returns the item if it was removed successfully.
    -- Returns false if the item was not found.
    -- Only the first item will be found.
    local pos = self:has(item)
    local r = self[pos]
    table.remove(self, pos)
    return r
end

function private.objects.list:has(item)
    -- Check the table to find the item.
    -- Return its position if found.
    -- Otherwise, return false.
    -- This is done using simple equality checks.
    -- Only the first item will be found.
    for i=1, #self do
        if self[i] == item then return i end
    end
    return false
end

function private.objects.list:merge(list)
    -- Merge this list with another
    -- The values from the argument list will be added to this list.
    -- The reverse will not happen.
    -- Values with keys will also be added,
    -- but the keys will be discarded.
    -- Returns the number of items added
    local n = 0
    for k, v in pairs(list) do
        self:add(v)
        n = n + 1
    end
    return n
end

function private.objects.list:clear()
    -- Clears the list, removing every item
    -- This is not the same as wipe(), as all
    -- non-list entries remain.
    while table.remove(self) ~= nil do end
end

--[[

A list object is an ordered structure of items. The prototype
provides functions for adding an item to the list, removing
an item from the list, and seeing if (and where) an item is
in the list. It is of course still a table, so looking up
a value by position can be done by list[pos].

]]

------------------------------------------------------------
--------------------- The queue class ----------------------
------------------------------------------------------------

private.objects.queue = {}
setmetatable(private.objects.queue, private.objects)

function private.objects.queue:new(tbl)
    -- Creates a new queue, using existing values if provided
    local r = private.objects.list(tbl)
    setmetatable(r, self)
    self.__index = self
    self.__call = getNew
    return r
end

function private.objects.queue:enqueue(item)
    -- Adds an item to the back of the queue.
    table.insert(self, item)
    return true
end

function private.objects.queue:dequeue(item)
    -- Removes and returns the item at the front of the queue.
    local r = self[1]
    table.remove(self, 1)
    return r
end

--[[

A queue object is a specialised list. It retains all functions
from the list object, but also gets the enqueue and dequeue
functions. Dequeueing removes and returns the first item at
the front of the queue, and enqueueing adds an item to the
back of the queue.

]]

------------------------------------------------------------
---------------------- The dict class ----------------------
------------------------------------------------------------

private.objects.dict = {}
setmetatable(private.objects.dict, private.objects)

function private.objects.dict:new(tbl)
    -- Creates a new dictionary, using existing values if provided
    local r = tbl or {}
    setmetatable(r, self)
    self.__index = self
    self.__call = getNew
    return r
end

function private.objects.dict:set(key, val)
    -- Adds the key-value pair to the dictionary if it doesn't exist.
    -- Updates it if it does.
    -- Returns the old value, if any.
    local r = self[key]
    self[key] = val
    return r
end

function private.objects.dict:unset(key)
    -- Removes the key-value pair from the dictionary.
    -- Returns the existing value, if any.
    -- Behaves identical to dict:set(key, nil).
    local r = self[key]
    self[key] = nil
    return r
end

function private.objects.dict:get(key)
    -- Returns the value of the key in the dictionary.
    return self[key]
end

--[[

A dictionary object is a list of key-value pairs. It's a very
simple class, as the functions are simply get-set-and-return
methods.

]]

------------------------------------------------------------
---------------------- The base class ----------------------
------------------------------------------------------------

local function frameEventHandler(self, event, addonName)
    self.__parent:event(event, addonName)
end

private.objects.base = {}
setmetatable(private.objects.base, private.objects)
function private.objects.base:new()
    local r = {}
    setmetatable(r, self)
    self.__index = self -- Enables inheritance
    self.__call = getNew -- Enables constructor
    r.__parent = self
    r.__eventFrame = CreateFrame("Frame")
    r.__eventFrame.__parent = r
    r.__eventFrame:SetScript("OnEvent", frameEventHandler)
    return r
end

function private.objects.base:init()
    -- Objects are expected to have an init() function
    return
end

local allowedTypes = private.objects.list({
    "number",
    "string",
    "boolean",
    "table",
})

function private.objects.base:serialise()
    -- Returns a table with all serialisable data in this table
    -- Override if a more specialised serialisation process is required
    local r = {}
    for k, v in pairs(self) do
        if allowedTypes:has(type(k)) and allowedTypes:has(type(v)) then
        --if isIn(type(k), allowedTypes) and isIn(type(v), allowedTypes) then
            -- Special handling for table keys and values
            -- If the table has a serialise function, use that
            -- Otherwise, serialise the table with this function
            if type(k) == "table" then
                k = k.serialise and k:serialise() or self.serialise(k)
            end
            if type(v) == "table" then
                v = v.serialise and v:serialise() or self.serialise(v)
            end
            r[k] = v
        end
    end
    return r
end

function private.objects.base:deserialise(tbl)
    -- Loads all values in the table into this object
    -- Override if a more specialised deserialisation process is required
    for k, v in pairs(tbl) do
        self[k] = v
    end
end

function private.objects.base:RegisterEvent(eventName)
    -- Registers an event with the object's event frame
    -- Will call self:event() when the event fires
    self.__eventFrame:RegisterEvent(eventName)
end

function private.objects.base:UnregisterEvent(eventName)
    -- Unregisters the event
    self.__eventFrame:UnregisterEvent(eventName)
end

function private.objects.base:event(event, addonName, ...)
    -- By default, simply calls the name of the called event as a function of the object.
    -- Override this to simply respond to any event, or create a function for the specific
    -- event to respond to that.
    if not self[event] then return end
    if type(self[event]) ~= "function" then return end
    return self[event](self, addonName, ...)
end

------------------------------------------------------------
----------------- Base class documentation -----------------
------------------------------------------------------------

--[[

The base class is used to create objects for any needs.
It provides a variety of useful functions.

Create new object:
    To create a new object, call base:new(), or just base()
    The new object will have access to all the features of the base class.
    If you override this function, make SURE to start off with this line:
        local r = class.__parent.new(self)
    where r is the new object. Do NOT use self in place of class.

Serialisation:
    The base class provides a basic serialisation function.
    It will simply return a table holding all key-value pairs where
    both key and value are either a number, a string, a boolean value,
    or a table. Tables will in turn be serialised themselves.
    Note that this function will NOT handle recursive tables. Running
    it on a table that, through any level of nesting, contains itself,
    will cause WoW to hang.
    
    The serialisation functions are intended to be overridden.

Event handling
    Event handling is also done through the base object. To respond to an
    event, you need to register the event using:
        object:RegisterEvent("eventName")
    Then, every time the event fires, the object:event() function will be
    called, with the event and the addon name as arguments. You need to
    override this function in order to do anything, however:
        object:event(eventName, addonName)

]]